<?php
// Validations

// Interpret intent of input value and attempt to convert into desired type
function normalize ($value, $type) {

	// Supported input/output formats
	$formats = array('boolean', 'string', 'integer', 'array');

	// Make strings safer
	if (is_string($value)) {
		$value = stripslashes($value);
	}

	// Single value of an array can be given as single input
	if ($type == 'array' and !is_array($value)) {
		$value = array($value);

	// Try to map to a boolean
	// FLAG not sure if it's a good idea to interpret strings like this
	} else if ($type == 'boolean' and !is_bool($value)) {
		$value = (empty($value) or (is_string($value) and in_array(strtolower($value), array('null', 'nil', 'false')))) ? false : true;

	// Try to map to an integer
	} else if ($type == 'integer' and !is_int($value)) {
		$value = intval($value);

	// Try to map to strings
	} else if ($type == 'string' and !is_string($value)) {

		// Implode an array
		if (is_array($value)) {
			$value = implode(', ', array_flatten($value));

		// Convert an integer to string
		} else if (is_int($value)) {
			$value = strval($value);

		}

	}

	return $value;
}



// Test an input value against a validation schema
// 
// FLAG
// Allow empty values? Currently, allows them as fallback but not in input for anything but integers
// Allow setting multiple alternative formats? Current normalization makes that useless.
// 
// NOTE
// If a proper default is missing, an attribute is assumed to be obligatory.
// Returns an array based on the model, overridden by the input values that pass the validation.
// Booleans always default to false and can't throw errors (if you need them to, you're doing it wrong)
function validate ($name, $schema, $input) {

	global $constants;

	$errors = array();
	$return = array();

	// Supported input formats
	$formats = array('boolean', 'string', 'integer', 'array', 'file');



	// Normalize prototype

	// Allow giving just a type, defaulting to others
	if (is_string($schema)) {
		$schema = array('type' => $schema);
	}

	// Prototype default is always false
	if (!isset($schema['default'])) {
		$schema['default'] = false;
	}

	// Defaulting to booleans on non-valid type
	if (!in_array($schema['type'], $formats) and $schema['type']) {
		$schema['type'] = $formats[0];
	}



	// Defaulting

	// No proper input value given
	if (empty($input) and $schema['type'] != 'integer') {

		// No fallback (empty fallbacks are valid)
		if ($schema['default'] === false and $schema['type'] != 'boolean') {
			$errors[] = '400 we need you to tell us what '.$name.' you want';

		// Fall back to default
		} else {
			$return = $schema['default'];

		}



	// Start evaluating input
	
	// An input value exists
	} else {

		// Normalize input value
		$input = normalize($input, $schema['type']);

		// Let's see if the type is acceptable
		if (gettype($input) == $schema['type']) {

			// Some other conditions for acceptable values given
			if (isset($schema['values']) and !empty($schema['values'])) {

				// A given set of acceptable values
				if (is_array($schema['values']) and !in_array($input, array_flatten($schema['values']))) {
					$errors[] = '400 please choose '.limplode(', ', $schema['values'], ' or ').' as '.$name;

				// Pattern name
				} else if (is_string($schema['values']) and !preg_match($schema['values'], $input)) {
					$errors[] = '400 you did not give a valid '.$name;
				}

				// FLAG add support for min/max integer values for integers

			}

			// The input seems acceptable
			if (empty($errors)) {
				$return = $input;
			}

		// File uploads
		} else if ($schema['type'] == 'file') {

			// Uploaded files are represented by a specific kind of array
			if (is_array($input) and isset($input['name'], $input['type'], $input['size'], $input['location']) and preg_match($constants['patterns']['content type'], $input['type'])) {

				// Additional restrictions
				if (isset($schema['values'])) {

					// Normalize file type restrictions
					$temp = array();
					foreach (normalize($schema['values'], 'array') as $value) {

						// Shorthand content type in schema
						if (isset($constants['content types'][$value])) {
							$temp[] = $constants['content types'][$value];

						// Full content type in schema
						} else if (preg_match($constants['patterns']['content type'], $value)) {
							$temp[] = $value;
						}

					}

					// Content type must match
					if (!in_array($input['type'], $temp)) {
						$errors[] = '400 only '.limplode(', ', $temp, ' and ').' for '.$name;
					}

				}

				// The input seems acceptable
				if (empty($errors)) {
					$return = $input;
				}

			// Not a file
			} else {
				$errors[] = '400 you must upload a file for '.$name;
			}

		// Wrong type
		} else {
			$errors[] = '400 '.$name.' must be '.$schema['type'];

		}

	}

	// Return errors in a retarded way
	if (!empty($errors)) {
		$return = array('errors' => $errors);
	}
	return $return;
}



// Run validate() for a set of values and schemas
function run_validations ($input, $schemas) {

	$errors = array();
	$return = array();

	// Iterate through supported parameters
	$i = 0;
	foreach ($schemas as $key => $value) {

		// Map shorthand values to associative values
		if (!isset($input[$key]) and isset($input[$i])) {
			$input[$key] = $input[$i];
		}

		// No value has been given for evaluation for this parameter, we'll try false
		if (!isset($input[$key])) {
			$input[$key] = false;
		}

		// Validate this input
		$temp = validate($key, $value, $input[$key]);
		if (is_array($temp) and isset($temp['errors']) and is_array($temp['errors'])) {
			$errors = array_merge($errors, $temp['errors']);
		} else {
			$return[$key] = $temp;
		}

		$i++;

	}

	// Return errors in a retarded way
	if (!empty($errors)) {
		$return = array('errors' => $errors);
	}

	return $return;
}

?>