/*
 * Baseline JavaScript library
 * Collection of structured and normalized JavaScript behavior for sane people
 * By Jerry Jäppinen
 * http://eiskis.net/

 * For docs & other info, check out livespec at
 * http://eiskis.net/proot/baseline/
 *
 * By Jerry Jäppinen
 * Released under the MIT license
 * eiskis@gmail.com
 * http://eiskis.net/
 * @Eiskis
 */
(function(window, undefined ) {

	// The actual Baseline library
	var bl = {

		// Some useful stuff to keep in mind
		conf: {
			version: '0.0.1',
			types: ['bool', 'int', 'float', 'str', 'array', 'hash', 'func']
		},

		// Type checks
		is: {

			// Unsets
			undef: function (value) {
				return value === undefined;
			},
			nul: function (value) {
				return value === null;
			},
			nan: function (value) {
				// NaN -_-
				return isNaN(value) && typeof value === 'number';
			},
			set: function (value) {
				return !(bl.is.undef(value) || bl.is.nul(value) || bl.is.nan(value));
			},

			// Emptiness
			empty: function (value) {
				var empty = false;

				if (
					(!bl.is.set(value)											) ||	// Unsets are always empty
					(bl.is.bool(value) && value === false						) ||	// Boolean must be true
					(bl.is.str(value) && bl.trim.empties(value).length === 0	) ||	// Strings must have something other than whitespaces
					(bl.is.float(value) && value === 0							) ||	// A number must not be 0
					(bl.is.array(value) && value.length === 0					) ||	// An array must have a length

					// A hash must have stuff
					(bl.is.hash(value) && (function (value) {for (var key in value) {return false;}return true;})(value))

				) {
					empty = true;
				}
				return empty;
			},

			// Type checks
			bool: function (value) {
				return (value === true || value === false);
			},
			int: function (value) {
				return bl.is.float(value) && value % 1 === 0;
			},
			float: function (value) {
				return typeof value === 'number';
			},
			str: function (value) {
				return typeof(value) === 'string';
			},
			array: function (value) {
				return value instanceof Array;
			},
			hash: function (value) {
				return $.isPlainObject(value);
			},
			func: function (value) {
				return $.isFunction(value);
			},
			collection: function (value) {
				return bl.is.array(value) || bl.is.hash(value);
			}

		},

		// NOTE we generate more are methods from is methods later
		are: {

			equal: function () {

				// Comparison of two items
				var _compare = function (a, b) {
					var result = false;

					// Start with type check
					if (bl.get.type(a) === bl.get.type(b)) {
						var type = bl.get.type(a);

						// Booleans, numbers and strings are easy
						if (bl.is.bool(a) || bl.is.float(a) || bl.is.str(a)) {
							result = (a === b);

						// Unsets of same type are always equal
						} else if (!bl.is.set(a)) {
							result = true;

						// Arrays are trickier
						} else if (bl.is.array(a)) {
							
							// Length must match
							// FLAG should I use bl methods in cases like this? They're slower but always consistent (maybe optimize production releases later)
							if (bl.get.length(a) === bl.get.length(b)) {

								// Evaluate reqursively
								result = true;
								for (var i = 0; i < a.length; i++) {
									if (!bl.are.equal(a[i], b[i])) {
										result = false;
										break;
									}
								}
							}

						// Hashes are the trickiest
						} else if (bl.is.hash(a)) {

							var aKeys = bl.list.keys(a);
							var bKeys = bl.list.keys(b);

							// Length must match, not order
							if (bl.get.length(aKeys) === bl.get.length(bKeys)) {
								result = true;

								// Key names and values must both match
								for (var j = 0; j < aKeys.length; j++) {
									var key = aKeys[j];
									if (!bl.is.set(bKeys[j]) || !bl.are.equal(a[key], b[key])) {
										result = false;
										break;
									}
								}

							}

						// Functions -_-
						} else if (bl.is.func(a)) {
							result = bl.to.str(a) === bl.to.str(b);
						}

					}

					return result;
				};

				var result = true;
				var length = arguments.length;

				// Compare each argument
				if (length >= 2) {
					for (var i = 1; i < length; i++) {
						if (!_compare(arguments[i-1], arguments[i])) {
							result = false;
							break;
						}
					}
				}

				return result;
			}

		},

		// Ensurances & conversion
		to: {

			// Return an empty value of the same type as the original
			empty: function (value) {
				if (bl.is.set(value)) {
					var result;
					if (bl.is.bool(value)) {
						result = false;
					} else if (bl.is.int(value)) {
						result = 0;
					} else if (bl.is.str(value)) {
						result = '';
					} else if (bl.is.array(value)) {
						result = [];
					} else if (bl.is.hash(value)) {
						result = {};
					} else if (bl.is.func(value)) {
						result = function () {};
					}
					return result;
				}
				return value;
			},

			// Type conversions
			bool: function (value) {
				var result = false;
				if (bl.is.set(value) && !bl.is.empty(value)) {
					result = true;
				}
				return result;
			},
			// FLAG: horribly unfinished
			int: function (value) {
				var result;
				var parsed = parseInt(value, 10);
				
				if (parsed) {
					result = parsed;
				} else {
					result = undefined;
				}

				return result;
			},
			// FLAG: horribly unfinished
			float: function (value) {
				var result;
				var parsed = parseFloat(value, 10);

				if (parsed) {
					result = parsed;
				} else {
					result = undefined;
				}

				return result;
			},
			str: function (value) {
				var result = '';

				if (bl.is.set(value)) {
					if (bl.is.str(value)) {
						result = value;

					// Basic things just get written into string
					} else if (bl.is.bool(value) || bl.is.int(value) || bl.is.float(value) || bl.is.func(value)) {
						result += value;

					// Keys in hash are dropped
					} else if (bl.is.array(value) && !bl.is.empty(value)) {
						result = value.join(', ');

					// Dump a hash
					} else if (bl.is.hash(value) && !bl.is.empty(value)) {
						result += '{' + "\n";

						var items = [];
						for (var key in value) {
							var item = value[key];
							items.push('    ' + key + ': ' + bl.to.str(item));
						}

						result += items.join(',' + "\n");

						result += "\n" + '}';
					}

				}

				return result;
			},
			array: function (value) {
				var result = [];

				if (bl.is.set(value) && !bl.is.empty(value)) {
					if (bl.is.array(value)) {
						result = value;

					// Basic things just get inserted as first value
					} else if (bl.is.bool(value) || bl.is.int(value) || bl.is.str(value) || bl.is.func(value)) {
						result.push(value);

					// Keys in hash are dropped
					} else if (bl.is.hash(value)) {
						for (var key in value) {
							result.push(value[key]);
						}

					}

				}

				return result;
			},
			hash: function (value, keys) {
				var result = {};

				// Allow setting new custom keys
				if (!bl.is.empty(keys)) {
					if (!bl.is.array(keys)) {
						keys = bl.to.array(keys);
					}
				} else {
					keys = [0];
				}

				if (bl.is.set(value) && !bl.is.empty(value)) {
					if (bl.is.hash(value)) {
						result = value;

					// Basic things just get inserted as first value
					} else if (bl.is.bool(value) || bl.is.int(value) || bl.is.str(value) || bl.is.func(value)) {
						result[keys[0]] = value;

					// Hash gets numerical keys
					} else if (bl.is.array(value)) {
						for (var i = 0; i < value.length; i++) {
							result[(keys[i] ? keys[i] : i)] = value[i];
						}

					}

				}

				return result;
			},
			func: function (value) {
				var result;

				if (bl.is.set(value) && !bl.is.empty(value)) {
					if (bl.is.func(value)) {
						result = value;

					// Basic things just get inserted as first value
					} else {
						result = function () {
							return value;
						};

					}

				} else {
					result = function () {};
				}

				return result;
			},
			collection: function (value) {

				// Hashes can stay hashes
				if (bl.is.hash(value)) {
					return bl.to.hash(value);

				// Others are converted to arrays
				} else {
					return bl.to.array(value);
				}
			}

		},

		// Getting imaginary parameters of an object
		get: {

			// Find out value's type
			type: function (value) {
				var types = bl.conf.types;
				
				// A value is there so we can find out its type
				if (bl.is.set(value)) {
					for (var i = 0; i < types.length; i++) {
						var type = types[i];
						if (bl.is[type](value)) {
							return type;
						}
					}

				// No value
				} else {
					return undefined;
				}
			},

			// FLAG: missing func
			length: function (value) {
				var length = 0;
				if (bl.is.array(value) || bl.is.str(value)) {
					length = value.length;
				} else if (bl.is.bool(value) && value === true) {
					length = 1;
				} else if (bl.is.hash(value)) {
					for (var key in value) {
						length++;
					}
				} else if (bl.is.int(value)) {
					length = bl.to.str(value).length;
				}
				return length;
			},

			// Output the value of an object (useful for cloning)
			value: function (original) {
				var result;

				// Arrays can be sliced
				if (bl.is.array(original)) {
					result = original.slice();

				// Objects need to be hacked
				} else if (bl.is.hash(original)) {
					result = $.extend({}, original);

				// For others, no relation is preserved
				} else {
					result = original;
				}

				return result;
			}
			
		},

		// Search for inclusions in collections
		has: {

			key: function (needle, haystack) {
				var result = false;
				if (bl.is.array(haystack) && bl.get.length(haystack) > needle) {
					result = true;
				} else if (bl.is.hash(haystack) && haystack[needle]) {
					result = true;
				}
				return result;
			},


			// Check for multiple needles as AND
			// FLAG only supports strings, returns false positives on arrays and hashes
			values: function (needles, haystack) {
				var result = true;
				needles = bl.to.array(needles);

				// Array haystack
				if (bl.is.array(haystack)) {

					// Hash haystack
				} else if (bl.is.hash(haystack)) {

				// Strings
				} else if (bl.is.str(haystack)) {
					
					for (var j = 0; j < needles.length; j++) {
						if (!bl.has.value(needles[j], haystack)) {
							result = false;
							break;
						}
					}

				}

				return result;
			},
				
			// FLAG support multiple needles as XOR
			value: function (needle, haystack) {
				var result = false;

				// Arrays
				if (bl.is.array(haystack)) {
					for (var i = 0; i < haystack.length; i++) {
						var hay = haystack[i];
						if (bl.are.equal(hay, needle)) {
							result = true;
							break;
						}
					}

				// Hashes
				} else if (bl.is.hash(haystack)) {
					for (var key in haystack) {
						if (bl.are.equal(haystack[key], needle)) {
							result = true;
							break;
						}
					}

				// Substrings
				} else if (bl.is.str(haystack)) {

					// Multiple needles (treated as XOR)
					if (bl.is.array(needle)) {
						for (var j = 0; j < needle.length; j++) {
							if (haystack.indexOf(needle[j]) >= 0) {
								result = true;
								break;
							}
						}

					} else {
						result = (haystack.indexOf(needle) < 0 ? false : true);
					}

				}

				return result;
			}

		},

		// Get arrays of shit
		list: {

			keys: function (value) {
				var result = [];

				// Array of index numbers
				if (bl.is.array(value)) {
					for (var i = 0; i < value.length-1; i++) {
						result.push(i);
					}

				// Array of the object's keys
				} else if (bl.is.hash(value)) {
					result = Object.keys(value);
				}

				return result.sort();
			},

			values: function (value) {
				return bl.to.array(value);
			},

			// List shared values
			shared: {

				values: function () {
					var type = bl.get.type(arguments[1]);

					// Prepare original deck of potentially shared values
					var result = bl.list.values(arguments[0]);

					// Iterate through all given parameters, starting from the second item
					if (bl.get.length(arguments) > 1) {
						for (var i = 1; i < arguments.length; i++) {
							var potential = bl.list.values(arguments[i]);

							// Compare values from the second item
							var newResult = [];
							for (var j = 0; j < result.length; j++) {
								var value = result[j];
								if (bl.has.value(value, potential)) {
									newResult.push(value);
								}
							}

							// Update deck for next iteration
							result = bl.get.value(newResult);

						}
					}

					return result;
				}

			}

		},

		// Pinpoint individual items from collections (return keys)
		find: {

			// NOTE this is pretty slow, and shouldn't be used unless really needed
			duplicates: function (collection) {
				var result = [];
				var i, j;

				// Arrays
				if (bl.is.array(collection)) {

					// Iterate through all values in collection
					for (i = 0; i < collection.length; i++) {

						// Only take each key in once
						if (!bl.has.value(i, result)) {

							// Compare to all upcoming values
							for (j = i+1; j < collection.length; j++) {
								if (bl.are.equal(collection[i], collection[j]) && (i != j) && !bl.has.value(j, result)) {
									result.push(j);
								}
							}

						}

					}

				// Hashes are treated a little differently
				} else if (bl.is.hash(collection)) {
					var keys = bl.list.keys(collection);

					for (i = 0; i < keys.length; i++) {
						var k = keys[i];
						if (!bl.has.value(k, result)) {

							// Compare to all upcoming values
							for (j = i+1; j < keys.length; j++) {
								var l = keys[j];
								if (bl.are.equal(collection[k], collection[l]) && (k != l) && !bl.has.value(l, result)) {
									result.push(l);
								}
							}

						}

					}

				}

				return result;
			},

			// Find items of a specific type
			// FLAG
			//    Maybe I should generate shortcuts for these, too?
			//    Parameter order inconsistent with other find methods (find.value(s) is using another method to achieve this)
			type: function (collection, types) {
				var result = [];

				// Support giving multiple types
				// FLAG I should try to do this more consistently, in a consistent manner, and find a generic way to do it (= internal method for evaluating parameters)
				if (arguments.length > 2) {
					types = bl.to.array(bl.remove.first(arguments));
				} else {
					types = bl.to.array(types);
				}

				// Arrays
				if (bl.is.array(collection)) {
					for (var i = 0; i < collection.length; i++) {
						for (var j = types.length - 1; j >= 0; j--) {
							if (bl.is[types[j]](collection[i])) {
								result.push(i);
								break;
							}
						}
					}

				// Hashes
				} else if (bl.is.hash(collection)) {
					for (var key in collection) {
						if (bl.has.value(bl.get.type(collection[key]), types)) {
							result.push(key);
						}
					}
				}

				return result;
			},

			// Get the keys of a value in a collection
			// FLAG should work for substrings
			values: function (needles, haystack) {
				var result = [];

				// Arrays
				if (bl.is.array(haystack)) {
					for (var i = 0; i < haystack.length; i++) {
						var hay = haystack[i];
						if (bl.has.value(hay, needles)) {
							result.push(i);
						}
					}

				// Hashes
				} else if (bl.is.hash(haystack)) {
					for (var key in haystack) {
						if (bl.has.value(haystack[key], needles)) {
							result.push(key);
						}
					}
				}

				return result;
			},
			value: function (needle, haystack) {
				return bl.find.values([needle], haystack);
			},

			// Getting the extremes
			first: {
				key: function (collection) {
					var result = bl.get.value(collection);
					
					if (bl.is.array(collection)) {
						result = (bl.is.empty(collection) ? undefined : 0);

					} else if (bl.is.hash(collection)) {
						var keys = Object.keys(collection);
						result = (bl.is.empty(collection) ? undefined : keys[0]);
					}
					return result;
				},

				// FLAG add value parameter for searching the first occurrence of something
				value: function (collection) {
					return collection[bl.find.first.key(collection)];
				}
				
			},

			last: {
				key: function (collection) {
					var result = bl.get.value(collection);
					if (bl.is.array(collection)) {
						result = collection.length-1;
					} else if (bl.is.hash(collection)) {
						var keys = bl.list.keys(collection).sort();
						result = keys[keys.length-1];
					}
					return result;
				},

				// FLAG add value parameter for searching the last occurrence of something
				value: function (collection) {
					var result = bl.get.value(collection);
					if (bl.is.array(collection) || bl.is.hash(collection)) {
						result = collection[bl.find.last.key(collection)];
					}
					return result;
				}
			}

		},

		// Negate remove
		preserve: {

			// Familiar shorthands (based on bl.find)
			duplicates: function (collection) {
				return bl.preserve.keys(collection, bl.find.duplicates.apply(bl.find, arguments));
			},
			type: function (collection, type) {
				return bl.preserve.keys(collection, bl.find.type.apply(bl.find, arguments));
			},
			values: function (needles, haystack) {
				return bl.preserve.keys(haystack, bl.find.values.apply(bl.find, arguments));
			},
			first: function (collection) {
				return bl.preserve.keys(collection, bl.find.first.key.apply(bl.find.first, arguments));
			},
			last: function (collection) {
				return bl.preserve.keys(collection, bl.find.last.key.apply(bl.find.last, arguments));
			},
			key: function (collection, key) {
				return collection[key];
			},
			
			// Preserve entries in a collection based on keys
			keys: function (collection, keys) {
				var result;
				var i;
				
				// Normalization & defaults
				keys = bl.to.array(keys);

				if (bl.is.array(collection)) {
					result = [];

					// Pick the marked items
					for (i = 0; i < collection.length; i++) {
						if (bl.has.value(i, keys)) {
							result.push(collection[i]);
						}
					}

				// Delete hash entries directly
				} else if (bl.is.hash(collection)) {
					result = {};

					// Copy marked items
					for (i = 0; i < keys.length; i++) {
						result[keys[i]] = collection[keys[i]];
					}
				}

				return result;
			}

		},

		// Remove items from collections
		remove: {

			// Trim aliases
			empties: function (value) {
				return bl.trim.empties(value);
			},

			// Familiar shorthands (based on bl.find)
			duplicates: function (collection) {
				return bl.remove.keys(collection, bl.find.duplicates.apply(bl.find, arguments));
			},
			type: function (collection, type) {
				return bl.remove.keys(collection, bl.find.type.apply(bl.find, arguments));
			},
			values: function (needles, haystack) {
				return bl.remove.keys(haystack, bl.find.values.apply(bl.find, arguments));
			},
			first: function (collection) {
				return bl.remove.keys(collection, bl.find.first.key.apply(bl.find.first, arguments));
			},
			last: function (collection) {
				return bl.remove.keys(collection, bl.find.last.key.apply(bl.find.last, arguments));
			},
			key: function (collection, key) {
				return bl.remove.keys(collection, [key]);
			},

			// Remove entries from a collection based on keys
			keys: function (collection, keys) {
				var result;
				var i;
				
				// Normalization & defaults
				keys = bl.to.array(keys);

				if (bl.is.array(collection)) {
					result = [];

					// Pick the items that aren't marked for death
					for (i = 0; i < collection.length; i++) {
						if (!bl.has.value(i, keys)) {
							result.push(collection[i]);
						}
					}

				} else if (bl.is.hash(collection)) {
					result = bl.get.value(collection);

					// Delete hash entries directly
					for (i = 0; i < keys.length; i++) {
						delete result[keys[i]];
					}
				}

				return result;
			}

		},

		// Strip all empty values and other excess shit off
		trim: {

			// Remove double whitespaces, turn tabs into single whitespaces, remove newlines
			whitespace: function (string) {
				return bl.trim.ends(string.replace(/\s+/g, ' '));
			},

			// Remove empty values and excess whitespace from string
			// FLAG should be bl.trim.all
			empties: function (collection) {
				var result = bl.get.value(collection);

				// Evaluate arrays recursively, remove all empty values
				// FLAG do I want this? Should everything touch only the topmost level (by default?), as that's what actually defines the object?
				if (bl.is.array(collection)) {
					result = [];
					for (var i = 0; i < collection.length; i++) {
						var arrayItem = bl.trim.empties(collection[i]);
						if (!bl.is.empty(arrayItem)) {
							// FLAG use bl.insert or something
							result.push(arrayItem);
						}
					}

				// Same for hashes
				} else if (bl.is.hash(collection)) {
					result = {};
					for (var key in collection) {
						var hashItem = bl.trim.empties(collection[key]);
						if (!bl.is.empty(hashItem)) {
							result[key] = (hashItem);
						}
					}
				}

				return result;
			},



			// Remove empty values and excess whitespace from the end
			start: function (value) {
				var result;

				// Strings
				if (bl.is.str(value)) {
					result = value.replace(/^[\s\xA0]+/, '');

				// Arrays
				} else if (bl.is.array(value)) {
					result = bl.get.value(value);
					while (bl.is.empty(result[0])) {
						result = bl.remove.first(result);
					}
				}

				return result;
			},

			// Remove empty values and excess whitespace from the end
			end: function (value) {
				var result;

				// Strings
				if (bl.is.str(value)) {
					result = value.replace(/[\s\xA0]+$/, '');

				// Arrays
				} else if (bl.is.array(value)) {
					result = bl.get.value(value);
					var i = value.length-1;
					while (bl.is.empty(result[i])) {
						result = bl.remove.last(result);
						i--;
					}
				}

				return result;
			},

			// Trim both ends
			ends: function (value) {
				// Prefer native trim for strings
				if (bl.is.str(value) && value.trim) {
					return value.trim();
				} else {
					return bl.trim.start(bl.trim.end(value));
				}
			}

		}

	};



	// Generate more functions

	// are methods from is methods
	(function () {
		for (var methodName in bl.is) {
			(function () {
				var method = methodName;
				bl.are[method] = function () {
					for (var i = 0; i < arguments.length; i++) {
						if (!bl.is[method](arguments[i])) {
							return false;
						}
					}
					return true;
				};
			})();
		}
	})();



	// Window scope

	// Expose bl to window scope
	window.bl = bl;

})(window);
